CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                      C
C  CARBON-RICH CIRCUMSTELLAR ENVELOPE CHEMICAL MODEL                   C
C     VERSION: 28 FEBRUARY 2012                                        C
C  BY M. A. CORDINER(1), T. J. MILLAR(2) AND A. J. MARKWICK(3)         C
C     (1) NASA GODDARD SPACE FLIGHT CENTER                             C
C     (2) QUEEN'S UNIVERSITY BELFAST                                   C
C     (3) THE UNIVERSITY OF MANCHESTER                                 C
C                                                                      C
C  SOLVES FOR CHEMICAL ABUNDANCES AS A FUNCTION OF RADIUS IN A UNIFORM C
C  SPHERICALLY-SYMMETRIC OUTFLOW WITH CONSTANT VELOCITY AND MASS-LOSS  C
C  RATE. HANDLES UP TO 995 SPECIES (INCLUDING UP TO 10 CONSERVED       C
C  SPECIES) LINKED BY UP TO 10000 BINARY REACTIONS                     C
C                                                                      C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

      PROGRAM CSE
      CHARACTER*12 SPECI(1000),PARENT(100)
      CHARACTER*11 JAC
      CHARACTER*500 FRATES,FSPECS,FOUTF,FOUTN
      DOUBLE PRECISION Y(1000),K(10000),X(10),TOTAL(10),COL(1000),
     *   B(1000,0:100),F(1000,0:100),PABUND(100),RADII(0:100),
     *   RWORK(2000000)
      DOUBLE PRECISION HNR,ACCR,TEMP,V,A_G,X_G,H2COL,AUV,MLOSS,MH,MU,KB,
     *   ALBEDO,ZETA,GAMCO,RINIT,RFINAL,RAD,R,ROUT,RTOL,ATOL,PI,CORATE,
     *   RANA,AUV_AV
      DOUBLE PRECISION GETHNR,GETTEM,GETH2,GETAUV,GETRAD,GETCOR
      INTEGER IWORK(1025),I,J,USPEC,URATES,UFRAC,UNUM,NSPEC,NCONS,
     *   NPAR,ICO,IRUN,IS,IF,LI,LTAG,ITASK,ISTATE,IOPT,LRW,LIW,MF,IANA,
     *   UANA
      EXTERNAL DIFFUN

      COMMON/BL1/ K,X,TOTAL,ACCR,HNR
      COMMON/BL2/ MLOSS,V
      COMMON/BL3/ Y,X_G,A_G,ZETA,ALBEDO,GAMCO,AUV_AV,ICO
      COMMON/BLC/ PI,MH,MU,KB


C  *****INITIALIZE BASIC MODEL PARAMETERS*******************************

C  DATA I/O UNIT NUMBERS
      DATA USPEC,URATES,UFRAC,UNUM,UANA,NEXTRA/1,2,3,4,5,5/
C  PHYSICAL CONSTANTS
      DATA PI,MH,MU,KB/3.1415927,1.6605E-24,2.2,1.3807E-16/

C  RATE FILE NAME
      FRATES = 'rate13.rates'
C  SPECIES FILE NAME
      FSPECS = 'rate12cse.specs'
C  FRACTIONAL ABUNDANCE OUTPUT FILE NAME
      FOUTF = 'csfrac_rate13.out'
C  NUMBER DENSITY OUTPUT FILE NAME
      FOUTN = 'csnum_rate13.out'

C  ANALYSE CHEMISTRY?
      IANA = 0
C  ANALYSIS RADIUS
      RANA = 2.0E16

C  INITIAL AND FINAL RADII (cm)
      RINIT = 1.0E15
      RFINAL = 7.0E17

C  OUTFLOW VELOCITY (cm/s)
      V = 14.5E5
C  MASS-LOSS RATE (M_sun/yr)
      MLOSS = 1.5E-5
C  GRAIN RADIUS (cm) (0.1 micron = 1.0E-5 cm)
      A_G = 1.0E-5
C  GRAIN NUMBER DENSITY/H2 (assuming gas/dust = 200, rho = 3.5 g/cm^3)
C  USED FOR H-ATOM ACCRETION CALCULATION
      X_G = 1.5E-12
C  GRAIN ALBEDO (FOR COSMIC-RAY-INDUCED PHOTON RATES)
      ALBEDO = 0.5
C  COSMIC-RAY IONISATION RATE SCALING FACTOR
      ZETA = 1.0
C  CO EXTINCTION COEFFICIENT
      GAMCO = 3.0
C  RATIO OF UV-TO-VISUAL EXTINCTION A(1000)/A(5500)
      AUV_AV = 4.65

C  OPEN RATE, SPECIES AND OUTPUT FILES
      OPEN(UNIT=USPEC, FILE=FSPECS, STATUS='OLD')
      OPEN(UNIT=UFRAC, FILE=FOUTF)
      OPEN(UNIT=UNUM, FILE=FOUTN)


C  *****READ IN CHEMICAL SPECIES****************************************

      WRITE(*,*) '...Reading species file...........'

      READ(USPEC,1)
 1    FORMAT(:)


C  INITIALISE NUMBER OF SPECIES
      NSPEC = 0
C  INITIALISE NUMBER OF CONSERVED SPECIES
      NCONS = 0
C  INITIALISE NUMBER OF PARENT SPECIES
      NPAR = 0

      I = 1

C  LOOP TO READ SPECIES NAMES
 100  READ(USPEC,2,END=101) J,SPECI(I)
 2       FORMAT(1X,I4,1X,A12)
C        PRINT*,J," ",SPECI(I)
C  DETERMINE INDEX OF CO
         IF(SPECI(I).EQ.'CO') ICO = I
         I=I+1
         IF (J.EQ.9999) THEN
            I=I-1
            NSPEC = I - 1
            GO TO 102
         ENDIF
      GO TO 100

C  LOOP TO READ CONSERVED SPECIES
 102  IF (SPECI(I).EQ.'Conserved:') THEN
         NCONS = 1
 103     READ(USPEC,3,END=101) J,SPECI(I),TOTAL(NCONS)
 3          FORMAT(1X,I4,1X,A12,1X,0PE8.2) 
C            PRINT*,J," ",SPECI(I),TOTAL(NCONS)
            I=I+1
            NCONS = NCONS + 1
            IF (J.EQ.9999) THEN
               I=I-1
               NCONS = NCONS - 2
               GO TO 104
            ENDIF
         GO TO 103
      ENDIF

C  LOOP TO READ PARENT ABUNDANCES
 104  IF (SPECI(I).EQ.'Parents:') THEN
         NPAR = 1
 105     READ(USPEC,*,END=101) PARENT(NPAR),PABUND(NPAR)
            NPAR = NPAR + 1
         GO TO 105
      ENDIF

 101  CONTINUE

      NPAR = NPAR - 1
      NTOT = NSPEC+NCONS

      WRITE(*,*) 'Species Read: ',NTOT
      WRITE(*,*) 'Conserved species: ',NCONS
      WRITE(*,*) 'Parent species: ',NPAR

C  CATCH ERRORS IN INPUT SPECIES
      IF(NSPEC.LT.1) CALL DIE
     *   ("No species defined in model -- check species input file    ")
      IF(NPAR.LT.1) CALL DIE
     *   ("No parents defined in model -- check species input file    ")


C  *****SET INITIAL PHYSICAL PARAMETERS*********************************

C  H2 NUMBER DENSITY
      HNR = GETHNR(RINIT)
C  KINETIC TEMPERATURE
      TEMP = GETTEM(RINIT)
C  H2 COLUMN DENSITY
      H2COL = GETH2(RINIT,HNR)
C  RADIAL UV EXTINCTION
      AUV = GETAUV(H2COL)
C  RADIATION FIELD STRENGTH
      RAD = GETRAD(AUV)

C  INITIALIZE ABUNDANCE ARRAY
      DO I = 1,NSPEC
         Y(I)= 0.0
      END DO

C  INITIALIZE 0TH ELEMENT OF OUTPUT DATA STORAGE ARRAYS
      DO I = 1,NTOT+NEXTRA
         B(I,0) = 0.0
         F(I,0) = 0.0
      END DO


C  *****SET INITIAL ABUNDANCES******************************************

      DO I = 1,NPAR
         DO J = 1,NTOT
            IF(SPECI(J).EQ.PARENT(I)) THEN
            Y(J) = PABUND(I)
C  LOAD FIRST ELEMENT OF OUTPUT ARRAYS
            B(J,0) = PABUND(I)
            F(J,0) = PABUND(I) * HNR
            WRITE(*,106) SPECI(J),PABUND(I)
 106        FORMAT(3X,A12,1PE8.2)
            ENDIF
         END DO
      END DO

C  INITIAL CO PHOTODISSOCIATION RATE
      CORATE = GETCOR(H2COL,Y(ICO),V,AUV)

C  LOAD 0TH ELEMENT OF CONSERVED SPECIES ARRAYS
      J = 1
      DO I = NSPEC+1,NTOT
         B(I,0) = TOTAL(J)
         F(I,0) = TOTAL(J) * HNR
         J = J+1
      END DO

C  PHYSICAL PARAMETER NAME STRINGS
      SPECI(NTOT+1) = 'n(H2)'
      SPECI(NTOT+2) = 'TEMP.'
      SPECI(NTOT+3) = 'A_V'
      SPECI(NTOT+4) = 'RAD. FIELD'
      SPECI(NTOT+5) = 'CO K(PHOT)'

C  LOAD 0TH ELEMENTS OF PHYSICAL PARAMETER ARRAYS
      B(NTOT+1,0) = HNR
      B(NTOT+2,0) = TEMP
      B(NTOT+3,0) = AUV/AUV_AV
      B(NTOT+4,0) = RAD
      B(NTOT+5,0) = CORATE


C  *****READ RATE FILE AND EXTRACT RATE DATA FOR EACH REACTION**********

      WRITE(*,*) '...Reading rate file..............'

      CALL READR(FRATES,URATES)

C  *****INITIALISE INTEGRATION STEP VARIABLES***************************

      R = RINIT
      ROUT = RINIT
      IRUN = 0
      RADII(IRUN) = ROUT


C  *****INITIALISE DVODE SOLVER VARIABLES*******************************

      LIW = NTOT + 30
      LRW = 22 + (9*NTOT) + (2*(NTOT**2))
      ITOL   = 1
      RTOL   = 1.0E-5
      ATOL   = 1.0E-20
      ITASK  = 1
      ISTATE = 1
      IOPT   = 0
      JAC    = 'DUMMYMATRIX'
      MF     = 22

      WRITE(*,*) '...Starting ODE solver DVODE......'


C  *****MAIN SOLVER LOOP************************************************

C  BEGIN LOOP AND INCREMENT RADIUS
 300  ROUT = 10.0**(LOG10(ROUT) + 0.05)
C  STEP NUMBER INCREMENT
      IRUN = IRUN + 1

C  CALL VARIABLE-COEFFICIENT ODE SOLVER (DVODE)
C  TO CALCULATE CHEMICAL ABUNDANCES (Y) WITH RESPECT TO RADIUS (ROUT)
C  BY INTEGRATION OF DY/DR VALUES CALCULATED BY SUBROUTINE 'DIFFUN'
      CALL DVODE (DIFFUN,NSPEC,Y,R,ROUT,ITOL,RTOL,ATOL,ITASK,
     &            ISTATE,IOPT,RWORK,LRW,IWORK,LIW,JAC,MF,RPAR,IPAR)

      WRITE(*,301) ROUT
 301  FORMAT(1X,"Radius: ",1PE8.2," cm")

C  *****LOAD OUTPUT ARRAYS**********************************************

C  PHYSICAL QUANTITIES
      HNR = GETHNR(ROUT)
      TEMP = GETTEM(ROUT)
      H2COL = GETH2(ROUT,HNR)
      AUV = GETAUV(H2COL)
      RAD = GETRAD(AUV)
      CORATE = GETCOR(H2COL,Y(ICO),V,AUV)
      RADII(IRUN) = ROUT

      B(NTOT+1,IRUN) = HNR
      B(NTOT+2,IRUN) = TEMP
      B(NTOT+3,IRUN) = AUV/AUV_AV
      B(NTOT+4,IRUN) = RAD
      B(NTOT+5,IRUN) = CORATE

C  CHEMICAL ABUNDANCES AND NUMBER DENSITIES
      DO I=1,NSPEC
         B(I,IRUN) = Y(I)
         F(I,IRUN) = Y(I)*HNR
      END DO

      DO I=1,NCONS
         B(NSPEC+I,IRUN) = X(I)
         F(NSPEC+I,IRUN) = X(I)*HNR
      END DO


C  *****CALL THE SUBROUTINE TO ANALYSE THE OUTPUT***********************

      IF(IANA.EQ.1.AND.ROUT.GE.RANA) THEN
      WRITE(*,*) '...Analysing chemistry............'
      CALL ANALYSE(SPECI,B,K,HNR,ROUT,IRUN,UANA,NTOT,FRATES,URATES)
      WRITE(*,*) '...Resuming model.................'
      IANA = 0
      ENDIF


C  TEST FOR END CONDITION

      IF(ROUT.LT.RFINAL) GO TO 300

      WRITE(*,310) FOUTF, FOUTN 
 310  FORMAT(1X,'...Writing results to',1X,A10,1X,'and',1X,A9)

C  *****CALCULATE COLUMN DENSITIES**************************************

      DO 400 I = 1,NTOT
         DO 401 J = 0,IRUN-1
            COL(I)=COL(I)+((RADII(J+1)-RADII(J))*(F(I,J)+
     *       (0.5*(F(I,J+1)-F(I,J)))))
 401     CONTINUE
 400  CONTINUE


C  *****OUTPUT ROUTINE**************************************************

      LI=NTOT+1
      IS=1
      IF=10
      WRITE(UFRAC,500)
 500  FORMAT(1X,' ** CALCULATED FRACTIONAL ABUNDANCES (/H2) ** '//)
      WRITE(UNUM,520)
 520  FORMAT(1X,' ** CALCULATED NUMBER DENSITIES (/CM^3) ** '//)
 501  WRITE(UFRAC,502)(SPECI(I),I=IS,IF)
 502  FORMAT(3X,'RADIUS',6X,10(1A12))
      WRITE(UNUM,502)(SPECI(I),I=IS,IF)
      LTAG=11
      DO 504 I=0,IRUN
         IF(I.EQ.LTAG) LTAG=LTAG+10
         WRITE(UFRAC,503) RADII(I),(B(J,I),J=IS,IF)
         WRITE(UNUM,503) RADII(I),(F(J,I),J=IS,IF)
 503  FORMAT(1X,11(1PE12.3E3))
 504  CONTINUE
      WRITE(UFRAC,502)(SPECI(I),I=IS,IF)
      WRITE(UNUM,502)(SPECI(I),I=IS,IF)
      WRITE(UFRAC,510)
      WRITE(UNUM,510)
      IS=IS+10
      IF=IF+10
      IF(IF.GT.(NTOT)) IF=NTOT
      IF(IS.GE.LI) GO TO 505
      GO TO 501
 
 505  WRITE(UFRAC,506)
      WRITE(UNUM,506)
 506  FORMAT(1X,' **  PHYSICAL PARAMETERS  **'//)
      WRITE(UFRAC,502)(SPECI(I),I=NTOT+1,NTOT+NEXTRA)
      WRITE(UNUM,502)(SPECI(I),I=NTOT+1,NTOT+NEXTRA)
      DO 507 I=0,IRUN
         WRITE(UFRAC,503) RADII(I),(B(J,I),J=NTOT+1,NTOT+NEXTRA)
         WRITE(UNUM,503) RADII(I),(B(J,I),J=NTOT+1,NTOT+NEXTRA)
 507   CONTINUE


C  OUTPUT COLUMN DENSITIES
      WRITE(UFRAC,510)
      WRITE(UNUM,510)
 510  FORMAT(//)
      WRITE(UFRAC,511)
      WRITE(UNUM,511)
 511  FORMAT(19X,'**  CALCULATED COLUMN DENSITIES  **'//)
      WRITE(UFRAC,512)(SPECI(I),COL(I),I=1,NTOT)
      WRITE(UNUM,512)(SPECI(I),COL(I),I=1,NTOT)
 512  FORMAT(3(1X,A12,'= ',1PE11.3,2X,:))

      CLOSE(UNIT=UFRAC)
      CLOSE(UNIT=UNUM)

      WRITE(*,*) "...DONE!!........................."

      STOP
      END
